/* eslint-disable @next/next/no-img-element */
/* eslint-disable @typescript-eslint/no-explicit-any */
import { readdir, readFile, stat } from "fs/promises";
import { glob } from "glob";
import Link from "next/link";
import { compileMDX } from "next-mdx-remote/rsc";
import path from "path";
import { JSXElementConstructor, ReactElement } from "react";
import remarkGfm from "remark-gfm";
import remarkUnwrapImages from "remark-unwrap-images";

import { NavigationItem } from "components/common/Navbar";

const DOCS_FOLDER = "../../docs";

import styles from "./layout.module.scss";

export type FrontMatter = {
    label?: string;
    icon?: JSX.Element;
    priority?: number;
};

// Get all files from docs folder and return an array of paths
export async function getAllowedMDXFiles(
    folder: string,
    recursive = true,
): Promise<
    {
        slug: string[];
        frontmatter: FrontMatter;
        content: ReactElement<any, string | JSXElementConstructor<any>>;
    }[]
> {
    //glob all mdx files
    if (recursive) {
        folder = `${folder}/**`;
    }
    const files = await glob(`${DOCS_FOLDER}/${folder}/*.mdx`);

    //map to array of paths
    const paths = await Promise.all(
        files.map(async (file) => {
            const path = file
                .replace(`${DOCS_FOLDER}/`, "")
                .replace(".mdx", "");

            //get frontmatter
            // Optionally provide a type for your frontmatter object
            const { content, frontmatter } = await compileMDX<FrontMatter>({
                source: await readFile(file, "utf-8"),
                options: { parseFrontmatter: true },
            });

            return { slug: path.split("/"), frontmatter, content };
        }),
    );

    return paths.sort((a, b) => {
        if (a.frontmatter.priority && b.frontmatter.priority) {
            return b.frontmatter.priority - a.frontmatter.priority;
        } else if (a.frontmatter.priority) {
            return -1;
        } else if (b.frontmatter.priority) {
            return 1;
        } else {
            return 0;
        }
    });
}

export async function getFoldersInDir(dirPath): Promise<string[]> {
    try {
        const files = await readdir(dirPath);
        const folders = [];

        for (const file of files) {
            const filePath = path.join(dirPath, file);
            const fileStat = await stat(filePath);

            if (fileStat.isDirectory()) {
                folders.push(file);
            }
        }

        return folders;
    } catch (error) {
        console.error("Error reading directory:", error);
        return [];
    }
}

export async function getSideBarItems(
    folder: string,
    recursive = true,
): Promise<NavigationItem[]> {
    const allowedFiles = await getAllowedMDXFiles(folder, recursive);

    return allowedFiles.map(({ slug, frontmatter }) => {
        return {
            label: frontmatter.label || "No label",
            href: `/docs/${slug.join("/")}`,
        };
    });
}

/**
 * Parses an MDX file and returns the compiled result.
 * @param source - The source code of the MDX file.
 * @param file_path - The file path of the MDX file.
 * @returns A promise that resolves to the compiled MDX result.
 */
export async function parseMDXFileFromSlug(slug: string[]) {
    let path = "docs/" + slug.slice(0, -1).join("/");
    if (!path.endsWith("/")) {
        path += "/";
    }

    const fp = "../../" + path + slug[slug.length - 1] + ".mdx";
    const source = await readFile(fp, "utf-8");
    return await compileMDX<FrontMatter>({
        source: source,
        options: {
            parseFrontmatter: true,
            mdxOptions: {
                remarkPlugins: [remarkGfm, remarkUnwrapImages],
            },
        },
        components: {
            a: (props) => {
                let { href } = props;
                if (href.endsWith(".mdx")) {
                    href = href.replace(".mdx", "");
                }

                return <Link href={href}>{props.children}</Link>;
            },
            img: async (props) => {
                // Load image if it ./images/ folder
                if (props.src.startsWith("./")) {
                    const img = await readFile(
                        `../../${path}/${props.src.slice(2)}`,
                    );

                    return (
                        <div className={styles.imgWrapper}>
                            <img
                                src={
                                    "data:image/jpeg;base64," +
                                    img.toString("base64")
                                }
                                alt={props.alt}
                            />
                        </div>
                    );
                }
            },
            table: (props) => {
                return (
                    <div className={styles.tableWrapper}>
                        <table className="table table-bordered table-striped">
                            {props.children}
                        </table>
                    </div>
                );
            },
            Alert: (props) => {
                return (
                    <div className={styles.alertWrapper}>
                        <div className={`alert alert-${props.variant}`}>
                            {props.children}
                        </div>
                    </div>
                );
            },
        },
    });
}
