/* eslint-disable @next/next/no-img-element */
import { useEffect } from "react";

import styles from "./Tables.module.scss";

/** Homebrew sortable table
 * also searchable see `searchTable`
 *
 */
export function Table({
    id,
    headers,
    rows,
    col_types,
    initial_sort = undefined,
}) {
    useEffect(() => {
        setup_sortable_tables(id);
    }, [id]);

    const thead_i = [];
    headers.forEach((header, i) => {
        let sorting = (
            <div className="table-sorting">
                <img src="img/caret-up.svg" alt="u" />
                <img src="img/caret-down.svg" alt="d" />
            </div>
        );
        if (col_types[i] === "ignore") {
            sorting = null;
        }
        thead_i.push(
            <th
                key={i}
                scope="col"
                data-type={col_types[i]}
                className={initial_sort == header ? "default-sort" : ""}
            >
                <div>
                    {header}
                    {sorting}
                </div>
            </th>
        );
    });
    const thead = (
        <thead>
            <tr>{thead_i}</tr>
        </thead>
    );

    const tbody_i = [];
    let c = 0;
    rows.forEach((row) => {
        const tr = [];
        row.forEach((cell) => {
            tr.push(<td key={c}>{cell}</td>);
            c++;
        });
        tbody_i.push(<tr key={c}>{tr}</tr>);
    });
    const tbody = <tbody>{tbody_i}</tbody>;

    return (
        <div className={styles.container}>
            <table
                id={id}
                className={"table table-striped table-hover " + styles.table}
            >
                {thead}
                {tbody}
            </table>
        </div>
    );
}

/**
 * Search a table
 */
export function searchTable(search_val: string, id_table: string) {
    // Declare variables
    let td, i, j, txtValue;
    const filter = search_val.toUpperCase();
    const table = document.getElementById(id_table);
    const tbody = table.querySelectorAll("tbody")[0];
    const tr = tbody.querySelectorAll("tr");
    // Loop through all table rows, and hide those who don't match the search query
    for (i = 0; i < tr.length; i++) {
        td = tr[i].querySelectorAll("td,th");
        for (j = 0; j < td.length; j++) {
            txtValue = td[j].textContent || td[j].innerText;
            if (txtValue.toUpperCase().indexOf(filter) > -1) {
                tr[i].style.display = "";
                break;
            } else {
                tr[i].style.display = "none";
            }
        }
    }
}

/** Sortable tables
 *
 *  Modified version of:
 *      https://htmldom.dev/sort-a-table-by-clicking-its-headers/
 */
function setup_sortable_tables(id) {
    const table = document.getElementById(id);

    if (table === null) {
        return;
    }

    const tableBody = table.querySelector("tbody");
    const tableHeader = table.querySelector("thead");
    const headers = tableHeader.querySelectorAll("th");
    const rows = tableBody.querySelectorAll("tr");

    // Track sort directions
    const directions = Array.from(headers).map(function () {
        return "unk";
    });

    // Transform the content of given cell in given column
    const transform = function (index, content) {
        // Get the data type of column
        const type = headers[index].getAttribute("data-type");
        switch (type) {
            case "ignore":
                return "";
            case "date":
                if (!content.trim()) {
                    return new Date(0);
                }
                return new Date(content);
            case "number":
                return parseFloat(content);
            case "string":
            default:
                return content;
        }
    };

    const sortColumn = function (index) {
        // Get the current direction
        const direction = directions[index] || "unk";

        // A factor based on the direction
        const multiplier = direction === "asc" ? 1 : -1;

        const newRows = Array.from(rows);

        newRows.sort(function (rowA, rowB) {
            const cellA = rowA.querySelectorAll("td,th")[index].innerHTML;
            const cellB = rowB.querySelectorAll("td,th")[index].innerHTML;

            const a = transform(index, cellA);
            const b = transform(index, cellB);

            switch (true) {
                case a > b:
                    return 1 * multiplier;
                case a < b:
                    return -1 * multiplier;
                case a === b:
                    return 0;
            }
        });

        // Remove old rows
        [].forEach.call(rows, function (row) {
            tableBody.removeChild(row);
        });

        // Set all directions to unk
        for (let i = 0; i < directions.length; i++) {
            directions[i] = "unk";
        }

        // Reverse the direction
        directions[index] = direction === "asc" ? "desc" : "asc";

        // Append new row
        newRows.forEach(function (newRow) {
            tableBody.appendChild(newRow);
        });
    };

    const changeIcons = function (index) {
        // Change image to filled version
        const row_headers = tableHeader.querySelectorAll("th");
        index = 0;
        for (const row_header of Array.from(row_headers)) {
            const direction = directions[index];
            const images = row_header.querySelectorAll("img");
            if (images.length < 2) {
                index++;
                continue;
            }
            switch (direction) {
                case "asc":
                    images[0].src = "img/caret-up-fill.svg";
                    images[1].src = "img/caret-down.svg";
                    break;
                case "desc":
                    images[0].src = "img/caret-up.svg";
                    images[1].src = "img/caret-down-fill.svg";
                    break;
                case "unk":
                    images[0].src = "img/caret-up.svg";
                    images[1].src = "img/caret-down.svg";
                    break;
                default:
                    break;
            }
            index++;
        }
    };

    [].forEach.call(headers, function (header, index) {
        header.addEventListener("click", function () {
            sortColumn(index);
            changeIcons(index);
        });
    });
    const clickEvent = new MouseEvent("click");
    table.getElementsByClassName("default-sort")[0]?.dispatchEvent(clickEvent);
}
