"use client";
import { logout } from "lib/actions/logout";
import { useActivePath } from "lib/hooks/useActivePath";
import useUser from "lib/hooks/useUser";
import Link from "next/link";
import { useRouter } from "next/navigation";
import { HTMLAttributes, ReactNode } from "react";
import { TbDots, TbDotsVertical } from "react-icons/tb";

import { Logo } from "./Header";
import { NavigationItem } from "./Navbar";
import {
    admin_svg,
    books_svg,
    logout_svg,
    person_svg,
    upload_svg,
} from "./svgs";

import styles from "./Sidebar.module.scss";

const sideBarItemsD: NavigationItem[] = [
    {
        label: "Books",
        href: "/books",
        icon: books_svg,
    },
    {
        label: "Account",
        href: "/account",
        icon: person_svg,
    },
    {
        label: "Upload",
        href: "/upload",
        icon: upload_svg,
    },
    {
        label: "Admin",
        href: "/admin",
        icon: admin_svg,
        condition: "admin",
    },
];

export function ContentWrapper({ children }) {
    return (
        <div className={styles.content}>
            <div>{children}</div>
        </div>
    );
}

export function Sidebar({
    sideBarItems = sideBarItemsD,
    logout = true,
    home = false,
}) {
    const isActivePath = useActivePath();
    return (
        <div className={styles.sidebar}>
            <div className={styles.branding}>
                <Logo />
            </div>
            <nav>
                <ul className={"nav nav-pills " + styles.tabs}>
                    {sideBarItems.map(
                        ({ href, label, icon, condition }, index) => (
                            <Tab_Link
                                key={index}
                                href={href}
                                active={isActivePath(href)}
                                img={icon}
                                condition={condition}
                            >
                                {label}
                            </Tab_Link>
                        ),
                    )}
                    {
                        // Only show logout button if logout is true
                        logout && <LogoutButton />
                    }
                    {home && (
                        <Tab_Link align="left" href="/" special img={books_svg}>
                            Home
                        </Tab_Link>
                    )}
                </ul>
            </nav>
        </div>
    );
}

export type NavigationFolder = {
    label: string;
    href: string;
    children: NavigationItem[];
};

export function SidebarFolders({
    folders,
    logout = true,
    home = false,
    back_link = undefined,
}: {
    folders: (NavigationFolder | NavigationItem)[];
    logout?: boolean;
    home?: boolean;
    back_link?: string;
}) {
    const isActivePath = useActivePath();
    return (
        <div className={styles.sidebar}>
            <div className={styles.branding}>
                <Logo />
            </div>
            <nav>
                <ul className={"nav nav-pills " + styles.tabs}>
                    {folders.map((folder, index) => {
                        if ("children" in folder) {
                            return (
                                <Tab_Folder
                                    key={index}
                                    folder={folder}
                                    isActivePath={isActivePath}
                                />
                            );
                        } else {
                            return (
                                <Tab_Link
                                    key={index}
                                    href={folder.href}
                                    active={isActivePath(folder.href)}
                                    img={folder.icon}
                                >
                                    {folder.label}
                                </Tab_Link>
                            );
                        }
                    })}
                    {
                        // Only show logout button if logout is true
                        logout && <LogoutButton />
                    }

                    {back_link && (
                        <Tab_Link
                            align="left"
                            href={back_link}
                            special
                            img={books_svg}
                        >
                            Back
                        </Tab_Link>
                    )}

                    {home && (
                        <Tab_Link align="left" href="/" special img={books_svg}>
                            Home
                        </Tab_Link>
                    )}
                </ul>
            </nav>
        </div>
    );
}

type Props = {
    children: ReactNode;
    href?: string;
    active?: boolean;
    special?: boolean;
    align?: "left" | "right";
    img?: ReactNode;
    condition?: "admin";
} & HTMLAttributes<HTMLDivElement | HTMLAnchorElement>;

function Tab_Link({
    children,
    href = undefined,
    active = false,
    special = false,
    align = "right",
    img = undefined,
    condition = undefined,
    ...props
}: Props) {
    const { isAdmin } = useUser();

    if (condition === "admin" && !isAdmin) {
        return;
    }

    if (!href) {
        return (
            <div
                className={styles.tab}
                data-align={align}
                data-active={active}
                data-special={special}
                {...props}
            >
                <span className={styles.img}>{img}</span>
                <span>{children}</span>
            </div>
        );
    } else {
        return (
            <Link
                href={href}
                className={styles.tab}
                data-align={align}
                data-active={active}
                data-special={special}
                {...props}
            >
                <span className={styles.img}>{img}</span>
                <span>{children}</span>
            </Link>
        );
    }
}

function Tab_Folder({
    folder,
    isActivePath,
}: {
    folder: NavigationFolder;
    isActivePath: (path: string) => boolean;
}) {
    const actives = folder.children.map((child) => isActivePath(child.href));
    const any_active = actives.some((active) => active);

    if (!any_active) {
        return (
            <Tab_Link
                href={folder.href}
                active={false}
                className={styles.tab + " " + styles.folder}
                img={<TbDotsVertical />}
            >
                {folder.label}
            </Tab_Link>
        );
    }

    return (
        <li>
            <Tab_Link
                active={true}
                className={styles.tab + " " + styles.folder}
                img={<TbDots />}
            >
                {folder.label}
            </Tab_Link>
            <ul className="mt-1">
                {folder.children.map((child, index) => {
                    return (
                        <Tab_Link
                            key={index}
                            href={child.href}
                            active={isActivePath(child.href)}
                            img={child.icon}
                        >
                            {child.label}
                        </Tab_Link>
                    );
                })}
            </ul>
        </li>
    );
}

function LogoutButton() {
    const { mutate } = useUser();
    const router = useRouter();
    return (
        <Tab_Link
            align="left"
            onClick={async () => {
                await logout();
                mutate(null);
                router.refresh();
            }}
            img={logout_svg}
            special
        >
            Logout
        </Tab_Link>
    );
}
