"use client";
import { useEffect, useRef, useState } from "react";
import { BsChevronDown, BsChevronUp } from "react-icons/bs";

import { ID } from "@snip/database/types";
import { BaseSnip } from "@snip/snips/general/base";
import { ImageSnip } from "@snip/snips/general/image";
import { localDateString } from "@snip/utils/dates";

import UploadDropzone from "./uploadDropzone";

import { useEditorContext } from "components/editor/context";

import styles from "./preview.module.scss";

/** The snip preview components
 * shows a preview of all queued snips
 * in the book. (previously named
 * snipsToPlace)
 *
 * It allows to pass a filter to only show
 * a subset of the snips.
 *
 */
export function PreviewSnips({
    snips,
    filter_ids,
    book_id,
}: {
    snips: BaseSnip[];
    filter_ids?: ID[];
    book_id: ID;
}) {
    const { activeSnipId, setActiveSnipId } = useEditorContext();

    return (
        <div className={styles.wrapper}>
            <UploadDropzone
                book_id={book_id}
                setActiveSnipId={setActiveSnipId}
            />
            {snips.map((snip, i) => {
                return (
                    <PreviewSingleSnip
                        snip={snip}
                        key={i}
                        hidden={
                            filter_ids ? !filter_ids.includes(snip.id) : false
                        }
                        onClick={(snip, e) => {
                            e.preventDefault();
                            setActiveSnipId((prev) => {
                                if (prev === snip.id) {
                                    return null;
                                }
                                return snip.id;
                            });
                        }}
                        active={snip.id === activeSnipId}
                    />
                );
            })}
        </div>
    );
}

/**
 * Renders a preview of a single snip.
 *
 * @param snipData - The data of the snip to be rendered.
 * @param active - Indicates whether the snip is active.
 * @param hidden - Indicates whether the snip is hidden.
 * @param onClick - The callback function to be called when the snip is clicked.
 */
function PreviewSingleSnip({
    snip,
    active = false,
    hidden = false,
    onClick,
}: {
    snip: BaseSnip;
    active?: boolean;
    hidden?: boolean;
    onClick?: (
        snip: BaseSnip,
        e: React.MouseEvent<HTMLDivElement | HTMLCanvasElement, MouseEvent>,
    ) => void;
}) {
    // Prevernts hydration mismatch
    const [infos, setInfos] = useState<{
        width: number;
        height: number;
        date: string;
    }>({
        width: 0,
        height: 0,
        date: "",
    });
    const canvasRef = useRef<HTMLCanvasElement>(null);
    const [collapsed, setCollapsed] = useState(false);

    //Render snip
    useEffect(() => {
        const ctx = canvasRef.current.getContext("2d");

        function renderSnip(snip: BaseSnip, ctx: RenderingContext2D) {
            if (ctx) {
                const trans = ctx.getTransform().inverse();
                const width = trans.a * 150;
                const height = trans.d * 150 * 1.41;

                ctx.clearRect(0, 0, width, height);
                //Scale canvas to fit snip
                const scale = Math.min(
                    150 / snip.width,
                    (150 * 1.41) / snip.height,
                );
                ctx.setTransform(scale, 0, 0, scale, 0, 0);
                ctx.translate(-snip.x, -snip.y);
                snip.render(ctx);
            }
        }

        function _helper() {
            renderSnip(snip, ctx);
            setInfos({
                width: snip.width,
                height: snip.height,
                date: snip.created ? localDateString(snip.created) : "",
            });
        }

        if (Object.hasOwn(snip, "ready")) {
            (snip as ImageSnip).ready?.then(_helper);
        } else {
            _helper();
        }

        return () => {
            ctx?.clearRect(0, 0, 20000, 20000);
        };
    }, [snip]);

    return (
        <div className={styles.snip} data-active={active} data-hidden={hidden}>
            <canvas
                ref={canvasRef}
                width={150}
                height={150 * 1.41}
                onClick={(e) => onClick?.(snip, e)}
            />
            <div
                className={styles.type}
                onClick={() => setCollapsed(!collapsed)}
            >
                <div className={styles.typeText}>
                    {snip.type}
                    {collapsed ? <BsChevronUp /> : <BsChevronDown />}
                </div>
                <InfoDropdown collapsed={collapsed} infos={infos} />
            </div>
        </div>
    );
}

function InfoDropdown({
    collapsed,
    infos,
}: {
    collapsed: boolean;
    infos: { width: number; height: number; date: string };
}) {
    return (
        <table
            style={{
                display: collapsed ? "block" : "none",
            }}
            className={styles.info}
        >
            <tbody>
                <tr>
                    <th>Created</th>
                    <td>{infos.date}</td>
                </tr>
                <tr>
                    <th>Width</th>
                    <td>{infos.width.toFixed()}</td>
                </tr>
                <tr>
                    <th>Height</th>
                    <td>{infos.height.toFixed()}</td>
                </tr>
            </tbody>
        </table>
    );
}
