"use client";
import { HTMLAttributes, useCallback, useEffect, useRef } from "react";

interface DragMoveProps extends HTMLAttributes<HTMLDivElement> {
    onDragMove: (e: PointerEvent, offset?: [number, number]) => void;
}

export default function DragMove({
    onDragMove,
    children,
    ...props
}: DragMoveProps) {
    const ref = useRef<HTMLDivElement>(null);
    const [handlePointerDown] = useDrag({ onDragMove });

    useEffect(() => {
        const el = ref.current;
        el.addEventListener("pointerdown", handlePointerDown);

        return () => {
            el.removeEventListener("pointerdown", handlePointerDown);
        };
    }, [handlePointerDown, onDragMove]);

    return (
        <div ref={ref} {...props}>
            {children}
        </div>
    );
}

export function useDrag({
    onDragMove,
    onPointerDown,
    onPointerUp,
}: {
    onDragMove: (e: PointerEvent, offset?: [number, number]) => void;
    onPointerDown?: (e: PointerEvent) => void;
    onPointerUp?: (e: PointerEvent) => void;
}) {
    const offset = useRef<[number, number]>([0, 0]);
    const isDragging = useRef(false);
    const handlePointerMove = useCallback(
        (e: PointerEvent) => {
            e.preventDefault();
            if (isDragging.current) onDragMove(e, offset.current);
        },
        [offset, onDragMove],
    );

    const handlePointerUp = useCallback(
        (e) => {
            isDragging.current = false;
            offset.current = [0, 0];
            if (onPointerUp) onPointerUp(e);
            // Remove event listener for pointer move
            window.removeEventListener("pointermove", handlePointerMove);
            window.removeEventListener("pointerup", handlePointerUp);
        },
        [handlePointerMove, onPointerUp],
    );

    const handlePointerDown = useCallback(
        (e) => {
            offset.current = [e.layerX, e.layerY];
            isDragging.current = true;
            if (onPointerDown) onPointerDown(e);
            // Add event listener for pointer move
            window.addEventListener("pointermove", handlePointerMove);
            window.addEventListener("pointerup", handlePointerUp);
        },
        [handlePointerMove, handlePointerUp, onPointerDown],
    );

    return [handlePointerDown, handlePointerUp, handlePointerMove];
}
