import {
    Dispatch,
    HTMLAttributes,
    ReactNode,
    SetStateAction,
    useCallback,
    useEffect,
} from "react";
import { Button } from "react-bootstrap";
import { TbArrowsMove } from "react-icons/tb";

import { BaseSnip } from "@snip/snips/general/base";
import { ImageSnip } from "@snip/snips/general/image";

import { Rect } from "../types";
import DragMove from "./drag";

import { useEditorContext } from "components/editor/context";
import { OverlayWorker } from "components/editor/worker/overlayWorker";

import styles from "./overlay.module.scss";

export function Overlay({
    children,
    className,
    ...props
}: HTMLAttributes<HTMLDivElement>) {
    const { overlayRef, overlayRect } = useEditorContext();
    let cN = styles.wrapper;
    if (className) {
        cN += " " + className;
    }

    return (
        <div
            ref={overlayRef}
            className={cN}
            style={{
                top: overlayRect.y,
                left: overlayRect.x,
                width: overlayRect.width,
                minHeight: overlayRect.height,
                transform: `rotate(${overlayRect.rot ?? 0}deg) ${
                    overlayRect.mirror ? "scaleX(-1)" : ""
                }`,
            }}
            {...props}
        >
            {children}
        </div>
    );
}

export function OverlayMover({
    onMove,
}: {
    onMove: Dispatch<SetStateAction<[number, number]>>;
}) {
    return (
        <DragMove
            className={styles.mover}
            onDragMove={(e) => {
                onMove([e.movementX, e.movementY]);
            }}
        >
            <TbArrowsMove />
        </DragMove>
    );
}

/**
 * Overlay outline
 *
 * @param type
 * @param margin
 * @returns
 */
export function OverlayOutline({
    children,
    inner,
    className,
    type = "round",
    color = "var(--bs-secondary)",
    style,
    ...props
}: HTMLAttributes<HTMLDivElement> & {
    margin?: string;
    type?: "round" | "square";
    inner?: ReactNode;
}) {
    let cN = styles.outline;
    if (className) {
        cN += " " + className;
    }

    style = style || {};

    if (color) {
        style["borderColor"] = color;
    }

    return (
        <>
            <div className={cN} {...props}>
                <div className={styles.accent} data-type={type} style={style}>
                    {inner}
                </div>
                {children}
            </div>
        </>
    );
}

export function OverlayContent({
    children,
    className,
    ...props
}: HTMLAttributes<HTMLDivElement>) {
    let cN = styles.content;
    if (className) {
        cN += " " + className;
    }

    return (
        <div className={cN} {...props}>
            {children}
        </div>
    );
}

export type Position =
    | "top"
    | "right"
    | "bottom"
    | "left"
    | "top-left"
    | "top-right"
    | "bottom-left"
    | "bottom-right";

export function OverlayHandles({
    positions,
    onMove,
    type,
}: {
    positions: Position[];
    onMove: (e: MouseEvent, pos: Position) => void;
    type?: "round" | "square";
    subRect?: Rect;
}) {
    return (
        <>
            {positions.map((pos) => {
                return (
                    <DragMove
                        key={pos}
                        onDragMove={(e) => {
                            onMove(e, pos);
                        }}
                        className={styles.handle}
                        data-position={pos}
                        data-type={type}
                    ></DragMove>
                );
            })}
        </>
    );
}

/**
 * Custom hook that adds a snip overlay to the editor.
 * @param snip The snip to be rendered in the overlay.
 * @param overlayWorker The overlay worker responsible for managing the overlay.
 */
export function useSnipOverlay(snip: BaseSnip, overlayWorker: OverlayWorker) {
    const render_snip = useCallback(
        (ctx) => {
            snip.overlay = true;
            snip?.render(ctx);
        },
        [snip],
    );

    useEffect(() => {
        if (!snip) return;

        overlayWorker.add(render_snip);
        (snip as ImageSnip).ready?.then(() => {
            overlayWorker.add(render_snip);
        });
        return () => {
            overlayWorker.remove(render_snip);
            overlayWorker.add(() => {});
        };
    }, [snip, overlayWorker, render_snip]);
}

/**
 * Placement button for any snip overlay that may include
 * placement functionality.
 *
 * @param children - The content to be rendered inside the button.
 * @param onClick - The event handler for the button's click event.
 * @param className - Additional CSS class names for the button.
 * @param props - Additional HTML attributes for the button element.
 * @returns The rendered overlay place button component.
 */
export function OverlayPlaceButton({
    onPlace,
    className,
    ...props
}: HTMLAttributes<HTMLButtonElement> & {
    onPlace: () => void;
    onPlaceCopy?: () => void;
}) {
    let cN = styles.place;
    if (className) {
        cN += " " + className;
    }

    return (
        <Button className={cN} variant="primary" onClick={onPlace} {...props}>
            Place
        </Button>
    );
}
