"use client";
import {
    createContext,
    useCallback,
    useContext,
    useEffect,
    useMemo,
    useState,
} from "react";

import { ImageSnip } from "@snip/snips/general/image";

import { Rect, ToolContextProps } from "../types";

import { useQueuedSnipsContext } from "components/context/socket/useQueuedSnipsContext";
import { useWorkerContext } from "components/context/viewer/useWorkerContext";
import { useEditorContext } from "components/editor/context";

export interface ImageEditorContextI {
    /** Sub positions divide an image into a sub rectangle
     *
     * in dom units
     */
    subRect: Rect;
    setSubRect: (f: (prev: Rect) => Rect) => void;

    /** Position of the overlay in
     * in dom units
     */
    imgRect: Rect;
    setImgRect: (f: (prev: Rect) => Rect) => void;

    /** Currently active snip */
    activeImgSnip: ImageSnip | null;
    resetImgSnip: () => void;
}

const ImageEditorContext = createContext<ImageEditorContextI>({
    subRect: {
        x: 0,
        y: 0,
        width: 10,
        height: 20,
    },
    setSubRect: () => {},
    imgRect: {
        x: 0,
        y: 0,
        width: 10,
        height: 20,
    },
    setImgRect: () => {},
    activeImgSnip: null,
    resetImgSnip: () => {},
});

export function useImageEditorContext() {
    return useContext(ImageEditorContext);
}

export function ImageEditorContextProvider({ children }: ToolContextProps) {
    const { overlayRect, setOverlayRect, pageWrappersRef } = useEditorContext();
    const { activeSnipId } = useEditorContext();
    const { snips } = useQueuedSnipsContext();
    const { overlayWorker } = useWorkerContext();

    const activeImgSnip: ImageSnip = useMemo(() => {
        const s = snips.get(activeSnipId);
        if (s instanceof ImageSnip) {
            return s;
        }
        return null;
    }, [activeSnipId, snips]);

    const [subRect, _setSubRect] = useState<Rect>({
        x: 0,
        y: 0,
        width: 10,
        height: 20,
    });

    const setImgRect = useCallback(
        (f: (a: Rect) => Rect) => {
            setOverlayRect((prev) => {
                const next = f(prev);

                // Check min/max values
                if (next.width < 5) next.width = 5;
                if (next.height < 5) next.height = 5;
                if (next.x < 0) next.x = 0;
                if (next.y < 0) next.y = 0;

                // Automatically resize image snip
                const idx = 0;
                const worker = overlayWorker.get(idx);
                if (!worker) {
                    return next;
                }

                const wrapper = pageWrappersRef.current.get(idx);
                const pageDiv = wrapper.children[0] as HTMLDivElement;

                const xy = worker.canvasCoords_2_dbCoords([
                    next.x - pageDiv.offsetLeft,
                    next.y - pageDiv.offsetTop,
                ]);
                const wh = worker.canvasCoords_2_dbCoords([
                    next.width,
                    next.height,
                ]);
                activeImgSnip.x = xy[0];
                activeImgSnip.y = xy[1];
                activeImgSnip.width = wh[0];
                activeImgSnip.height = wh[1];

                /** Also resize the subrect
                 * to match the new image size
                 */
                const scale_x = activeImgSnip.width / activeImgSnip.img_width;
                const scale_y = activeImgSnip.height / activeImgSnip.img_height;
                // Scale subrect
                const sxy = worker.dbCoords_2_canvasCoords([
                    activeImgSnip.sx,
                    activeImgSnip.sy,
                ]);
                const swh = worker.dbCoords_2_canvasCoords([
                    activeImgSnip.swidth,
                    activeImgSnip.sheight,
                ]);

                _setSubRect({
                    x: sxy[0] * scale_x,
                    y: sxy[1] * scale_y,
                    width: swh[0] * scale_x,
                    height: swh[1] * scale_y,
                });

                worker.rerender();
                return next;
            });
        },
        [setOverlayRect, overlayWorker, pageWrappersRef, activeImgSnip],
    );

    const setSubRect = useCallback(
        (f: (a: Rect) => Rect) => {
            _setSubRect((prev) => {
                const next = f(prev);

                // Automatically apply changes to the image snip
                const idx = 0;
                const worker = overlayWorker.get(idx);
                if (!worker) {
                    return next;
                }

                const sub_xy = worker.canvasCoords_2_dbCoords([next.x, next.y]);
                const sub_wh = worker.canvasCoords_2_dbCoords([
                    next.width,
                    next.height,
                ]);

                const scale_x = activeImgSnip.width / activeImgSnip.img_width;
                const scale_y = activeImgSnip.height / activeImgSnip.img_height;
                activeImgSnip.sx = sub_xy[0] / scale_x;
                activeImgSnip.sy = sub_xy[1] / scale_y;
                activeImgSnip.swidth = sub_wh[0] / scale_x;
                activeImgSnip.sheight = sub_wh[1] / scale_y;
                worker.rerender();

                return next;
            });
        },
        [activeImgSnip, overlayWorker],
    );

    /** Whenever the snip changes reset the
     * overlay rect to the snips width and height
     * also reset the subrect.
     */
    useEffect(() => {
        if (!activeImgSnip || !(activeImgSnip instanceof ImageSnip)) return;

        const trans_wh = overlayWorker
            .get(0)
            .dbCoords_2_canvasCoords([
                activeImgSnip.width,
                activeImgSnip.height,
            ]);

        setImgRect((prev) => {
            return {
                ...prev,
                width: trans_wh[0],
                height: trans_wh[1],
            };
        });
        setSubRect(() => {
            return {
                x: 0,
                y: 0,
                width: trans_wh[0],
                height: trans_wh[1],
            };
        });
    }, [setSubRect, activeImgSnip, overlayWorker, setOverlayRect, setImgRect]);

    function resetImgSnip() {
        setImgRect((rect) => {
            return {
                ...rect,
                width: activeImgSnip.img_width ?? 0,
                height: activeImgSnip.img_height ?? 0,
                rot: 0,
            };
        });
        setSubRect(() => {
            return {
                x: 0,
                y: 0,
                width: activeImgSnip.img_width ?? 0,
                height: activeImgSnip.img_height ?? 0,
            };
        });
    }

    return (
        <ImageEditorContext.Provider
            value={{
                subRect,
                setSubRect,
                imgRect: overlayRect,
                setImgRect,
                activeImgSnip,
                resetImgSnip,
            }}
        >
            {children}
        </ImageEditorContext.Provider>
    );
}
