import { useEffect, useState } from "react";
import { TbPointer, TbPointerFilled } from "react-icons/tb";

import { get_snip_from_data } from "@snip/snips";
import { BaseSnip } from "@snip/snips/general/base";

import { Help, HelpDescription, HelpHeader } from "../common/help";
import { Toggle } from "../common/toggle";
import { ToolComponents, ToolToggleProps } from "../types";
import SnipToDom from "./snipToDom";

import { useWorkerContext } from "components/context/viewer/useWorkerContext";
import { useEditorContext } from "components/editor/context";
import {
    type InteractionEvent,
    PageWorker,
} from "components/editor/worker/pageWorker";

function InteractionToggle(props: ToolToggleProps) {
    const icon = props.isActive ? <TbPointerFilled /> : <TbPointer />;

    return (
        <Toggle label="Interact" icon={icon} expandable={false} {...props} />
    );
}

function InteractionHelp() {
    return (
        <Help>
            <HelpHeader>
                <span>Interaction</span>
                <TbPointer />
            </HelpHeader>
            <HelpDescription>
                <p>
                    Using this tool you may interact with the page. You may
                    select text, follow links or interact with interactive
                    snips.
                </p>
            </HelpDescription>
        </Help>
    );
}

/** The interaction overlay functions very differently to other
 * overlays!
 *
 * Instead of being used to place a snip on the page. It is used to
 * create html elements for text snips or other interactive components
 * on the page.
 *
 * Basically whenever a user clicks or hovers a snip the overlay will
 * create a html element for the snip and display it on the page.
 * This allows the user to interact with the snip in a more natural way.
 *
 */
function InteractionOverlay() {
    const { worker } = useWorkerContext();
    const { pageWrappersRef } = useEditorContext();
    const [init, setInit] = useState(false);
    const [snip, setSnip] = useState<BaseSnip | null>(null);

    /** Setup worker interaction
     * events when the overlay is mounted
     */
    useEffect(() => {
        const w = worker.get(0) as PageWorker;
        if (!w || init) {
            return;
        }
        w.setupInteraction().then(() => {
            setInit(true);
        });

        return () => {
            if (!init) return;
            w.teardownInteraction();
        };
    }, [init, pageWrappersRef, worker]);

    useEffect(() => {
        if (!init) return;
        const page = pageWrappersRef.current?.get(0)
            ?.children?.[0] as HTMLElement;
        const w = worker.get(0) as PageWorker;
        if (!page || !w) return;

        const remove = w.attachElementEvents(page);

        return () => {
            remove();
        };
    }, [init, pageWrappersRef, worker]);

    /** Register enter and leave events and
     * create elements for text snips
     */
    useEffect(() => {
        function enterSnip(
            e: InteractionEvent<"pointerenter" | "pointerdown">,
        ) {
            if (e.snipData.type == "image") return;
            const snip = get_snip_from_data(e.snipData);

            console.log("snip", snip);
            setSnip(snip);
        }

        const w = worker.get(0) as PageWorker;
        if (!w) return;

        w.interactions.addEventListener("pointerenter", enterSnip);
        w.interactions.addEventListener("pointerdown", enterSnip);

        return () => {
            w.interactions.removeEventListener("pointerenter", enterSnip);
            w.interactions.removeEventListener("pointerdown", enterSnip);
        };
    }, [worker]);

    return <SnipToDom snip={snip} />;
}

const InteractionTool: ToolComponents = {
    name: "interaction" as const,
    Toggle: InteractionToggle,
    Help: InteractionHelp,
    Overlay: InteractionOverlay,
};

export default InteractionTool;
