import { useEffect, useRef, useState } from "react";

import { BaseSnip } from "@snip/snips/general/base";
import { TextSnipLegacy } from "@snip/snips/general/legacy/text";
import { TextSnip } from "@snip/snips/general/text";

import { useMoveToolContext } from "../move/context";

import { useWorkerContext } from "components/context/viewer/useWorkerContext";
import { useEditorContext } from "components/editor/context";

import styles from "./interaction.module.scss";

export default function SnipToDom({ snip }: { snip: BaseSnip }) {
    if (snip instanceof TextSnip) {
        return <TextSnipAsFC snip={snip} />;
    }
    if (snip instanceof TextSnipLegacy) {
        return <TextSnipLegacyAsFC snip={snip} />;
    }
    // TODO: add overlays for other snips
    // e.g. links
    // add right click menu for snips to view some information
    return null;
}

/** Generate a text snip as a functional component
 * which can be used in the overlay
 */
function TextSnipAsFC({ snip }: { snip: TextSnip }) {
    const { zoom } = useMoveToolContext();
    const { overlayWorker } = useWorkerContext();
    const { pageWrappersRef } = useEditorContext();
    const [parsedText, setParsedText] = useState<string | null>(null);
    const ref = useRef<HTMLTextAreaElement>(null);

    useEffect(() => {
        const page = pageWrappersRef.current?.get(0)
            ?.children?.[0] as HTMLElement;
        const ele = ref.current;
        const z = zoom.get(0);
        if (!page || !ele || !z) return;

        // Set positions
        const xy = overlayWorker
            .get(0)
            .dbCoords_2_canvasCoords([snip.x, snip.y]);
        ele.style.left = page.offsetLeft + xy[0] + "px";
        // No idea where the offset comes from most likely css related
        const offset = (snip.canvasFont.fontSize * z) / 12;
        ele.style.top = page.offsetTop + xy[1] - offset + "px";

        //Set styles
        ele.style.fontFamily = snip.canvasFont.fontFamily;
        ele.style.fontSize = snip.canvasFont.fontSize * z + "px";
        ele.style.lineHeight =
            snip.canvasFont.lineHeight * snip.canvasFont.fontSize * z + "px";

        ele.style.width = snip.width * z + "px";
        ele.style.height = snip.height * z + "px";

        // Parse and set text
        let text = snip.text;
        text = text
            .replaceAll(/\t/g, "        ")
            .replaceAll(/\n/g, "\\n")
            .replaceAll(/\r/g, "")
            .replaceAll(/\\n/g, String.fromCharCode(13, 10));
        setParsedText(text);
    }, [overlayWorker, pageWrappersRef, snip, zoom]);

    return (
        <textarea
            ref={ref}
            className={styles.textSnip}
            value={parsedText}
            readOnly
        />
    );
}

function TextSnipLegacyAsFC({ snip }: { snip: TextSnipLegacy }) {
    const { zoom } = useMoveToolContext();
    const { overlayWorker } = useWorkerContext();
    const { pageWrappersRef } = useEditorContext();
    const [parsedText, setParsedText] = useState<string | null>(null);
    const ref = useRef<HTMLTextAreaElement>(null);

    useEffect(() => {
        const page = pageWrappersRef.current?.get(0)
            ?.children?.[0] as HTMLElement;
        const ele = ref.current;
        let z = zoom.get(0);
        if (!page || !ele || !z) return;
        z = z * 2.35;
        // Set positions
        const xy = overlayWorker
            .get(0)
            .dbCoords_2_canvasCoords([snip.x, snip.y]);
        ele.style.left = page.offsetLeft + xy[0] + "px";
        // No idea where the offset comes from most likely css related
        const offset = (snip.fontSize * z) / 6;
        ele.style.top = page.offsetTop + xy[1] - offset + "px";

        //Set styles
        ele.style.fontFamily = snip.fontFamily;
        ele.style.fontSize = snip.fontSize * z + "px";
        //ele.style.lineHeight = snip.lineHeight + "px";

        ele.style.width = snip.width * z + "px";
        ele.style.height = snip.height * z + "px";
        ele.style.lineHeight = snip.lineHeight * z + "px";
        // Parse and set text
        let text = snip.text;
        text = text
            .replaceAll(/\t/g, "        ")
            .replaceAll(/\n/g, "\\n")
            .replaceAll(/\r/g, "")
            .replaceAll(/\\n/g, String.fromCharCode(13, 10));
        setParsedText(text);
    }, [overlayWorker, pageWrappersRef, snip, zoom]);

    return (
        <textarea
            ref={ref}
            className={styles.textSnip}
            value={parsedText}
            readOnly
        />
    );
}
