import { createContext, useCallback, useContext, useMemo } from "react";

import { BaseSnip } from "@snip/snips/general/base";

import { ToolContextProps } from "../types";

import { useQueuedSnipsContext } from "components/context/socket/useQueuedSnipsContext";
import { useWorkerContext } from "components/context/viewer/useWorkerContext";
import { useEditorContext } from "components/editor/context";
import { OverlayWorker } from "components/editor/worker/overlayWorker";
import { PageWorker } from "components/editor/worker/pageWorker";

export interface PlacementContextI {
    activeSnip: BaseSnip | null;
    currentOverlayWorker: OverlayWorker | null;
    setPos: (x: number, y: number) => void;
    setPosDB: (x: number, y: number) => void;
    posDB: [number, number];
    setRot: (rot: number, mirror?: boolean) => void;
    align: (args: {
        x?: "left" | "center" | "right";
        y?: "top" | "center" | "bottom";
    }) => void;
    move: (movX: number, movY: number) => void;
    placeOnPage: () => Promise<void>;
}

const PlacementContext = createContext<PlacementContextI>({
    activeSnip: null,
    currentOverlayWorker: null,
    posDB: [0, 0],
    setPos: () => {},
    setPosDB: () => {},
    setRot: () => {},
    align: () => {},
    move: () => {},
    placeOnPage: async () => {},
});

export function PlacementContextProvider({ children }: ToolContextProps) {
    const {
        overlayRect,
        setOverlayRect,
        pageWrappersRef,
        activeSnipId,
        overlayRef,
        setActiveSnipId,
    } = useEditorContext();
    const { snips } = useQueuedSnipsContext();
    const { overlayWorker, worker } = useWorkerContext();

    const activeSnip: BaseSnip | null = useMemo(() => {
        return snips.get(activeSnipId);
    }, [activeSnipId, snips]);

    // The page on which the snip is placed is always the first page
    const PLACE_PAGE = 0;

    const currentOverlayWorker: OverlayWorker | null = useMemo(() => {
        return overlayWorker.get(PLACE_PAGE);
    }, [overlayWorker]);

    /** Update the position of the
     * currently active snip
     *
     * This also updates the overlay rect.
     *
     * x,y in dom units
     */
    const setPos = useCallback(
        (x: number, y: number) => {
            setOverlayRect((prev) => {
                // Check min/max bounds
                // Update activeSnip
                const wrapper = pageWrappersRef.current.get(PLACE_PAGE);
                const pageDiv = wrapper.children[0] as HTMLDivElement;

                const xy = currentOverlayWorker.canvasCoords_2_dbCoords([
                    x - pageDiv.offsetLeft,
                    y - pageDiv.offsetTop,
                ]);
                activeSnip.x = xy[0];
                activeSnip.y = xy[1];
                currentOverlayWorker.rerender();
                // Update overlay rect
                return { ...prev, x, y };
            });
        },
        [activeSnip, currentOverlayWorker, pageWrappersRef, setOverlayRect],
    );

    const move = useCallback(
        (movX: number, movY: number) => {
            setOverlayRect((prev) => {
                const new_x = prev.x + movX;
                const new_y = prev.y + movY;

                // Check min/max bounds
                // Update activeSnip
                const wrapper = pageWrappersRef.current.get(0);
                const pageDiv = wrapper.children[0] as HTMLDivElement;

                const xy = currentOverlayWorker.canvasCoords_2_dbCoords([
                    new_x - pageDiv.offsetLeft,
                    new_y - pageDiv.offsetTop,
                ]);
                activeSnip.x = xy[0];
                activeSnip.y = xy[1];
                currentOverlayWorker.rerender();

                return {
                    ...prev,
                    x: new_x,
                    y: new_y,
                };
            });
        },
        [activeSnip, currentOverlayWorker, pageWrappersRef, setOverlayRect],
    );

    /**
     * Sets the position of the active snip in the overlay.
     *
     * x,y in db units
     */
    const setPosDB = useCallback(
        (x: number, y: number) => {
            setOverlayRect((prev) => {
                // Check min/max bounds
                // Update activeSnip
                activeSnip.x = x;
                activeSnip.y = y;
                currentOverlayWorker.rerender();
                // Update overlay rect
                const xy = currentOverlayWorker.dbCoords_2_canvasCoords([x, y]);
                const pageDiv = pageWrappersRef.current.get(0)
                    .children[0] as HTMLDivElement;

                return {
                    ...prev,
                    x: xy[0] + pageDiv.offsetLeft,
                    y: xy[1] + pageDiv.offsetTop,
                };
            });
        },
        [activeSnip, currentOverlayWorker, pageWrappersRef, setOverlayRect],
    );

    /** Update the rotation
     * and mirror of the currently
     * active snip
     *
     * This also updates the overlay rect
     */
    const setRot = useCallback(
        (rot: number, mirror?: boolean) => {
            setOverlayRect((prev) => {
                // Update activeSnip
                activeSnip.rot = rot;
                activeSnip.mirror = mirror ?? false;
                currentOverlayWorker.rerender();
                // Update overlay rect
                return { ...prev, rot, mirror };
            });
        },
        [activeSnip, currentOverlayWorker, setOverlayRect],
    );

    const posDB = useMemo(() => {
        const wrapper = pageWrappersRef.current?.get(0);
        if (!wrapper) return [NaN, NaN] as [number, number];
        const pageDiv = wrapper.children[0] as HTMLDivElement;
        const xy = currentOverlayWorker.canvasCoords_2_dbCoords([
            overlayRect.x - pageDiv.offsetLeft,
            overlayRect.y - pageDiv.offsetTop,
        ]);
        return xy as [number, number];
    }, [overlayRect, pageWrappersRef, currentOverlayWorker]);

    const align = useCallback(
        ({
            x = "none",
            y = "none",
        }: {
            x?: "left" | "center" | "right" | "none";
            y?: "top" | "center" | "bottom" | "none";
        }) => {
            const wrapper = pageWrappersRef.current.get(0);
            const page = wrapper.children[0] as HTMLDivElement;
            const overlay = overlayRef.current;

            if (!wrapper || !overlay) return;

            let newX = overlayRect.x;
            let newY = overlayRect.y;

            const padding = 10;
            switch (y) {
                case "top":
                    newY = page.offsetTop + padding;
                    break;
                case "bottom":
                    console.log("page", page);
                    newY =
                        page.offsetTop +
                        page.scrollHeight -
                        overlay.clientHeight -
                        padding;
                    break;
                case "center":
                    newY =
                        page.offsetTop +
                        page.scrollHeight / 2 -
                        overlay.clientHeight / 2;
                    break;
            }

            switch (x) {
                case "left":
                    newX = page.offsetLeft + padding;
                    break;
                case "right":
                    newX =
                        page.offsetLeft +
                        page.offsetWidth -
                        overlay.clientWidth -
                        padding;
                    break;
                case "center":
                    newX =
                        page.offsetLeft +
                        page.scrollWidth / 2 -
                        overlay.clientWidth / 2;
                    break;
            }

            // apply to active snip
            const xy = currentOverlayWorker.canvasCoords_2_dbCoords([
                newX - page.offsetLeft,
                newY - page.offsetTop,
            ]);
            activeSnip.x = xy[0];
            activeSnip.y = xy[1];
            currentOverlayWorker.rerender();

            setOverlayRect((prev) => {
                return {
                    ...prev,
                    x: newX,
                    y: newY,
                };
            });
        },
        [
            activeSnip,
            currentOverlayWorker,
            overlayRect.x,
            overlayRect.y,
            overlayRef,
            pageWrappersRef,
            setOverlayRect,
        ],
    );

    /** Places the currently
     * active snip on the page
     * using the worker.
     *
     * I.e. moves the snip from the overlay
     * to the page.
     */
    async function placeOnPage() {
        if (!activeSnip) return;

        const w = worker.get(PLACE_PAGE) as PageWorker;
        await w?.placeOnPage(activeSnip.to_data());
        setActiveSnipId(null);
    }

    return (
        <PlacementContext.Provider
            value={{
                activeSnip,
                currentOverlayWorker,
                posDB,
                setPos,
                setPosDB,
                setRot,
                align,
                move,
                placeOnPage,
            }}
        >
            {children}
        </PlacementContext.Provider>
    );
}

export function usePlacementContext() {
    return useContext(PlacementContext);
}
