/**
 * README:
 * -------
 *
 * The overlay "worker" is used in the main thread to render
 * the overlay canvas. This is done to avoid the overhead of
 * creating a webworker for the overlay.
 *
 * Thus it is not a real worker but a facade to the overlay
 * rendering functions.
 */
import { Socket } from "socket.io-client";

import type { PageData } from "@snip/database/types";
import { BaseData, BaseView, SnipData } from "@snip/snips/general/base";
import { get_snip_from_data } from "@snip/snips/get_snip_from_data";

import { RenderWorker } from "./abc";

const debug = (...args: unknown[]) => {
    console.debug("[OverlayWorker]", ...args);
};

export class OverlayWorker extends RenderWorker {
    private ctx: CanvasRenderingContext2D;
    private stack = new Array<(ctx: CanvasRenderingContext2D) => void>();
    private socket: Socket;

    // Canvas size in db coordinates
    private max: [number, number] = [1400, 2000];

    constructor(socket: Socket) {
        super();
        this.socket = socket;
    }

    public async setupCanvas(canvas: HTMLCanvasElement) {
        this.canvas = canvas;
        this.ctx = canvas.getContext("2d", {
            desynchronized: true,
        });
        debug("Setup");
    }

    public async setupPage(_page: PageData) {
        // Check if page is already setup
        if (this.current_page_id === _page.id || !this.ctx) return;
        if (this.current_page_id) {
            // Leave room
            this.socket?.emit("page:leave", this.current_page_id);
        }
        this.current_page_id = _page.id;

        // Clear stack from previews
        for (const renderfn of this.preview2render.values()) {
            this.remove(renderfn);
        }
        this.preview2render.clear();
        this.resize(this.canvas.width, this.canvas.height);

        if (!this.socket) return;

        // Join room
        return new Promise<void>((resolve, reject) => {
            this.socket.emit(
                "page:join",
                this.current_page_id,
                (success: boolean) => {
                    if (success) {
                        resolve();
                    } else {
                        reject();
                    }
                },
            );
        });
    }

    public sendPreviewData(snipData: SnipData<BaseData, BaseView>) {
        if (!this.socket) return;
        this.socket.volatile.emit("snip:preview", snipData);
    }

    private preview2render = new Map<
        number,
        (ctx: CanvasRenderingContext2D) => void
    >();
    public receivePreviewData(snipData: SnipData<BaseData, BaseView>) {
        if (!this.current_page_id || snipData.page_id !== this.current_page_id)
            return;

        // Check if snip is already rendered and remove it
        if (this.preview2render.has(snipData.id)) {
            const renderfn = this.preview2render.get(snipData.id);
            this.remove(renderfn);
        }

        // Add to stack
        const snip = get_snip_from_data(snipData);
        this.preview2render.set(snipData.id, (ctx) => {
            snip.render(ctx);
        });
        this.add(this.preview2render.get(snipData.id));
    }

    public resize(
        width: number,
        height: number,
        devicePixelRatio = 1,
        by_width = false,
    ) {
        if (!this.canvas) {
            // Wait for canvas to be setup
            // this fixes a small bug where the canvas is not yet
            // created
            return setTimeout(() => this.resize(width, height), 1000);
        }

        // without zoom the canvas is 1400x2000
        this.canvas.width = width * devicePixelRatio;
        this.canvas.height = height * devicePixelRatio;

        const max_db = [1400, 2000];
        let scale = height / max_db[1];
        if (by_width) {
            scale = width / max_db[0];
        }

        this.ctx.setTransform(scale, 0, 0, scale, 0, 0);
        this.max = this.canvasCoords_2_dbCoords([
            this.canvas.width,
            this.canvas.height,
        ]) as [number, number];
        this.rerender();
    }

    public rerender() {
        if (this.stack.length === 0) return;

        this.ctx.clearRect(0, 0, this.max[0], this.max[1]);
        this.stack.forEach((fn) => fn(this.ctx));
    }

    public add(fn: (ctx: CanvasRenderingContext2D) => void) {
        this.stack.push(fn);
        this.rerender();
        debug("Added", this.stack.length);
    }
    public remove = (fn: (ctx: CanvasRenderingContext2D) => void) => {
        const index = this.stack.indexOf(fn);
        if (index !== -1) {
            this.stack.splice(index, 1);
        }
        //this.render();
        debug("Removed", this.stack.length);
    };

    public terminate() {
        this.ctx.clearRect(0, 0, 1400, 2000);
        debug("Terminated");
    }

    /** Transform a point from canvas coordinates to
     * transformed coordinates (1400x2000).
     */
    public canvasCoords_2_dbCoords(xy: [number, number]) {
        const scale = this.canvas.height / 2000;

        const x = xy[0] / scale;
        const y = xy[1] / scale;
        return [x, y] as [number, number];
    }
    public dbCoords_2_canvasCoords(xy: [number, number]) {
        const scale = this.canvas.height / 2000;
        const x = scale * xy[0];
        const y = scale * xy[1];
        return [x, y];
    }

    attachElementEvents(): () => void {
        //Does nothing for the overlay worker
        return () => {};
    }

    public getScale() {
        if (!this.canvas) return 1;
        return this.canvas.height / 2000;
    }
}
