import { PageData } from "@snip/database/types";
import { ArraySnip } from "@snip/snips/general/array";
import { BaseSnip } from "@snip/snips/general/base";
import { ImageSnip } from "@snip/snips/general/image";

import { render_background_by_id } from "./backgrounds";

/** A static page. A page which can render a predefined pageData
 * no dynamic content is used. See Page in fullstack for dynamic content
 * using a socket connection.
 */
export class LocalPage implements Required<PageData> {
    public id: number;
    public book_id: number;
    public page_number: number;
    public readonly created: Date;
    public readonly last_updated: Date | null;
    public background_type_id: number | null;

    protected snips: BaseSnip[];

    /** Create a new page object
     *
     * This allows your to access the render function of the page object.
     *
     * @param id The id of the page
     * @param book_id The id of the book
     * @param page_number The page number
     * @param created The date the page was created
     * @param last_updated The date the page was last updated
     * @param background_type_id The id of the background type
     * @param snips The snips on the page can be none if not loaded
     */
    constructor({
        id,
        book_id,
        page_number,
        created = new Date(),
        last_updated = new Date(),
        background_type_id = null,
        snips = [],
    }: PageData & { snips?: BaseSnip[] }) {
        this.id = id;
        this.book_id = book_id;
        this.page_number = page_number;
        this.last_updated = last_updated;
        this.created = created;
        this.background_type_id = background_type_id;
        this.snips = snips;
    }

    static from_data(data: PageData): LocalPage {
        return new LocalPage({
            ...data,
        });
    }

    public to_data(): Required<PageData> {
        return {
            id: this.id,
            book_id: this.book_id,
            page_number: this.page_number,
            created: this.created,
            last_updated: this.last_updated,
            background_type_id: this.background_type_id,
        };
    }

    public set_snips(snips: BaseSnip[]): void {
        this.snips = snips;
    }

    public add_snip(snip: BaseSnip): void {
        /** If array snip remove all
         * nested snips and add them to the arraysnip
         */
        if (snip instanceof ArraySnip) {
            snip.snips.forEach((s) => {
                this.remove_snip(s.id);
            });
        }

        /** Get all array snips and check if they contain the snip
         */
        const array_snips = this.snips.filter(
            (s) => s instanceof ArraySnip,
        ) as ArraySnip[];
        for (const array_snip of array_snips) {
            for (let s of array_snip.snips) {
                if (s.id === snip.id) {
                    s = snip;
                    return;
                }
            }
        }

        this.snips.push(snip);
    }

    public get_snip_at(coords: [number, number]): BaseSnip | null {
        // Check if snips are loaded
        if (this.snips === undefined) {
            return null;
        }
        // Check if snip is at coords
        let snips = [];
        if (this.snips.length > 25) {
            // If more than 25 snips filter out doodle snips
            snips = this.snips.filter((snip) => snip.type != "doodle");
        } else {
            snips = this.snips;
        }
        // Reverse order as the last snip is on top
        snips = snips.reverse();

        const snip = snips.find((s) => {
            const poly = s.polygon;
            return ptinpoly(coords[0], coords[1], poly);
        });
        if (snip) {
            return snip;
        }
        return null;
    }

    public get_snip_by_id(snip_id: number): BaseSnip | null {
        const snip = this.snips.find((s) => s.id === snip_id);
        if (snip) {
            return snip;
        }
        return null;
    }

    public update_snip(snip: BaseSnip): void {
        const index = this.snips.findIndex((s) => s.id === snip.id);
        if (index !== -1) {
            this.snips[index] = snip;
        } else {
            this.add_snip(snip);
        }
    }

    public remove_snip(snip_id: number): boolean {
        const index = this.snips.findIndex((s) => s.id === snip_id);
        if (index !== -1) {
            this.snips.splice(index, 1);
            return true;
        }
        return false;
    }

    public render(
        ctx: CanvasRenderingContext2D | OffscreenCanvasRenderingContext2D,
    ): void {
        // await a load of all images

        // Render background
        render_background_by_id(this.background_type_id, ctx);

        const sorted_snips = this.snips.sort(
            (a, b) => a.created.getTime() - b.created.getTime(),
        );

        for (const snip of sorted_snips) {
            snip.render(ctx);
        }
    }
}

// poly is an array of [ x, y ] elements
function ptinpoly(x: number, y: number, poly: [number, number][]) {
    let c = false;
    for (let l = poly.length, i = 0, j = l - 1; i < l; j = i++) {
        const xj = poly[j][0],
            yj = poly[j][1],
            xi = poly[i][0],
            yi = poly[i][1];
        const where = (yi - yj) * (x - xi) - (xi - xj) * (y - yi);
        if (yj < yi) {
            if (y >= yj && y < yi) {
                if (where == 0) return true; // point on the line
                if (where > 0) {
                    if (y == yj) {
                        // ray intersects vertex
                        if (y > poly[j == 0 ? l - 1 : j - 1][1]) {
                            c = !c;
                        }
                    } else {
                        c = !c;
                    }
                }
            }
        } else if (yi < yj) {
            if (y > yi && y <= yj) {
                if (where == 0) return true; // point on the line
                if (where < 0) {
                    if (y == yj) {
                        // ray intersects vertex
                        if (y < poly[j == 0 ? l - 1 : j - 1][1]) {
                            c = !c;
                        }
                    } else {
                        c = !c;
                    }
                }
            }
        } else if (y == yi && ((x >= xj && x <= xi) || (x >= xi && x <= xj))) {
            return true; // point on horizontal edge
        }
    }
    return c;
}
