import { createCanvas, DOMMatrix, Image, registerFont } from "canvas";
import { glob } from "fast-glob";
import { createReadStream, createWriteStream, existsSync, mkdirSync } from "fs";
import path from "path";
import sharp from "sharp";
import { Readable } from "stream";
import { pipeline } from "stream/promises";

import service from "@snip/database";
import { FONT_URLS } from "@snip/database/fonts/internal";
import { get_snip_from_data } from "@snip/snips";
import {
    UnknownNamespaceError,
    UnknownSnipTypeError,
} from "@snip/snips/errors";
import { ArraySnip } from "@snip/snips/general/array";
import { DummySnip } from "@snip/snips/general/dummy";
import { ImageSnip } from "@snip/snips/general/image";
import { ImageSnipLegacy } from "@snip/snips/general/legacy/image";
import { TextSnip } from "@snip/snips/general/text";
import { config } from "@snip/utils/config";

import { LocalPage } from "../localPage";
import { log } from "../logging";
import { type RenderParams } from "./renderParams";

const get_bitmap = async function () {
    this.bitmap = new Image();
    const promise = new Promise((resolve, reject) => {
        this.bitmap.onload = () => {
            resolve(true);
        };
        this.bitmap.onerror = (error) => {
            reject(error);
        };
    });

    // Webp support for node-canvas kinda sucks
    if (this.blob.mime === "image/webp") {
        const png_buffer = await sharp(Buffer.from(this.blob.data, "base64"))
            .toFormat("png")
            .toBuffer();
        this.bitmap.src = `data:image/png;base64,${png_buffer.toString("base64")}`;
    } else {
        this.bitmap.src = `data:${this.blob.mime};base64,${this.blob.data}`;
    }
    await promise;

    this.img_width = this.bitmap.width;
    this.img_height = this.bitmap.height;
    // Overwrite default values if not set
    this._swidth = this.swidth || this.img_width;
    this._sheight = this.sheight || this.img_height;
    this._width = this.width || this.img_width;
    this._height = this.height || this.img_height;
    return true;
};

ImageSnipLegacy.prototype.get_bitmap = get_bitmap;
ImageSnip.prototype.get_bitmap = get_bitmap;

TextSnip.prototype.get_overlap_index = function (str, maxWidth, font) {
    if (maxWidth == -1) {
        return undefined;
    }
    const canvas = createCanvas(1, 1);
    const ctx = canvas.getContext("2d");
    ctx.font = font;

    let width = 0;
    for (let c = 0; c < str.length; c++) {
        const char = str.charAt(c);
        width += ctx.measureText(char).width;
        if (width > maxWidth) {
            return c;
        }
    }
    return undefined;
};

// We need to load fonts for the page
// this can be problematic for tsc
// so we have to use a config path
for (const [name, family] of FONT_URLS) {
    registerFont(path.join(__dirname, "../../../../assets/fonts/", name), {
        family: family,
    });
}

/** Partially public render function
 * allows to render a page by its id
 */
export default async function renderPage(
    params: RenderParams,
): Promise<Readable | true> {
    /** Check if file exists already everything with a higher format
     * than the requested one will be rendered
     *
     * /renders/{page_id}/{width}.{format}
     */
    const dir = `${config.RENDER.DIR}/${params.page_id}`;

    let transform_webp = false;
    if (params.format === "webp") {
        transform_webp = true;
        params.format = "jpeg";
    }

    let stream;
    if (!params.force_render) {
        stream = await pageFromFile(
            params.page_id,
            params.width,
            params.format,
        );
    }

    let stream_promise: Promise<true>;
    if (stream == undefined) {
        stream = await pageFromDB(params.page_id, params.width, params.format);
        // Write the stream to the file system
        if (!existsSync(dir)) {
            mkdirSync(dir, {
                recursive: true,
            });
        }
        stream_promise = pipeline(
            stream,
            createWriteStream(`${dir}/${params.width}.${params.format}`),
            {
                end: false,
            },
        ).then(() => true);
        //log("Rendered page", params.page_id, params.width, params.format);
    } else {
        //log("Found page", params.page_id, params.width, params.format);
        stream_promise = Promise.resolve(true);
    }

    if (!params.return_stream) {
        return await stream_promise;
    }

    // Transform to webp if requested
    if (transform_webp) {
        params.format = "webp";
        const webp_transform = sharp().webp();
        stream = stream.pipe(webp_transform);
    }

    return stream;
}

/** Helper function to find a rendered file
 * if it exists
 *
 * @param page_id the id of the page
 * @param width the width of the file
 * @param format the format of the file
 * @returns a readable stream if the file exists undefined otherwise
 */
async function pageFromFile(
    page_id: number,
    width: number,
    format: string,
): Promise<Readable | undefined> {
    const dir = `${config.RENDER.DIR}/${page_id}`;
    const files = await glob(`${dir}/*.${format}`);
    for (const file of files) {
        const [fileWidth, fileFormat] = file.split("/").pop()!.split(".");
        if (width === parseInt(fileWidth)) {
            return createReadStream(file);
        }
    }
}

/**
 * Renders a page from the database and returns the stream of the rendered page.
 * @param page_id - The ID of the page to render.
 * @param width - The width of the rendered page.
 * @param format - The format of the rendered page, which can be "pdf", "jpeg", or "png".
 * @returns A Promise that resolves to a Readable stream of the rendered page, or undefined if the page data is not found.
 */
async function pageFromDB(
    page_id: number,
    width: number,
    format: "pdf" | "jpeg" | "png",
): Promise<Readable | undefined> {
    /** Get page data
     * including all snips
     */
    const pageData = await service.page.getById(page_id);
    const page = LocalPage.from_data(pageData);
    const snipsData = await service.snip.getByPageId(page_id);
    const snips = snipsData.map((snipData) => {
        try {
            return get_snip_from_data(snipData);
        } catch (e) {
            if (e instanceof UnknownSnipTypeError) {
                log("Unknown snip type", e.message);
            } else if (e instanceof UnknownNamespaceError) {
                log("Unknown namespace", e.message);
            } else if (e instanceof Error) {
                log("Unknown error", e.message);
            } else {
                log("Unknown error", e.toString());
            }
        }
        return new DummySnip(snipData);
    });

    // Prepare snip images
    const promises: Promise<boolean>[] = [];
    for (const snip of snips) {
        if (snip instanceof ImageSnip || snip instanceof ImageSnipLegacy) {
            promises.push(snip.ready);
        }
        if (snip instanceof ArraySnip) {
            for (const child_snip of snip.snips) {
                if (
                    child_snip instanceof ImageSnip ||
                    child_snip instanceof ImageSnipLegacy
                ) {
                    promises.push(child_snip.get_bitmap());
                }
            }
        }
    }
    await Promise.all(promises);
    page.set_snips(snips);

    /** Render the page
     * write the stream to the file system
     * and return the stream
     */
    const stream = render(page, width, format);
    return stream;
}

/** Helper function to render a local page
 *  expects a fully initialized local page!
 */
export function render(
    localPage: LocalPage,
    width: number,
    format: "pdf" | "jpeg" | "png",
): Readable {
    const canvas = createCanvas(
        width,
        (width * 2000) / 1400,
        format != "pdf" ? undefined : "pdf",
    );

    const ctx = canvas.getContext("2d");
    const transform = ctx.getTransform().invertSelf();
    const t_width =
        transform.a * canvas.width + transform.b * canvas.height + transform.e;
    const t_height =
        transform.c * canvas.width + transform.d * canvas.height + transform.f;
    ctx.fillStyle = "white";
    ctx.fillRect(0, 0, t_width, t_height);

    const scale = canvas.height / 2000;
    const t = new DOMMatrix([scale, 0, 0, scale, 0, 0]);
    ctx.setTransform(t);
    localPage.render(ctx as unknown as CanvasRenderingContext2D);

    switch (format) {
        case "pdf":
            return canvas.createPDFStream();
        case "jpeg":
            return canvas.createPNGStream();
        case "png":
            return canvas.createJPEGStream({ quality: 0.6 });
    }
}
