import SQLService from "@snip/database";
import { pool_data } from "@snip/database/sql.connection";
import { BaseData, BaseView, SnipData } from "@snip/snips/general/base";

import { type SocketData } from "../../types";
import { logWS } from "../../utils/logging";
import { triggerRender } from "../../utils/triggerRender";

export function setupSnipPrefix({
    socket,
    book_nsp,
    server,
    ...props
}: SocketData) {
    socket.on(
        "snip:insert",
        async (
            snip_data: SnipData<BaseData, BaseView>,
            callback: (id: number | null) => void,
        ) => {
            // Check if user has permission to write to the page
            if (!(socket.perms.pWrite === true)) {
                return socket.emit(
                    "error",
                    "You do not have access to write to this book. Could be finished!",
                );
            }

            return await insert(
                { socket, book_nsp, server, ...props },
                snip_data,
                callback,
            );
        },
    );

    socket.on(
        "snip:update",
        async (
            snip_data: SnipData<BaseData, BaseView>,
            callback: (id: number | null) => void,
        ) => {
            // Check if user has permission to write to the page
            if (!(socket.perms.pWrite === true)) {
                return socket.emit(
                    "error",
                    "You do not have access to write to this book. Could be finished!",
                );
            }

            return await update(
                { socket, book_nsp, server, ...props },
                snip_data,
                callback,
            );
        },
    );

    // Remove snip from db
    socket.on(
        "snip:remove",
        async (snip_id: number, callback: (success: boolean) => void) => {
            // Check if user has permission to delete from the page
            if (!(socket.perms.pDelete === true)) {
                callback?.(false);
                return socket.emit(
                    "error",
                    "You do not have access to delete from this book.",
                );
            }
            logWS(socket, "snip:remove " + snip_id);
            await SQLService.snip.delete(snip_id);
            callback?.(true);
            // Remove placed snip
            book_nsp.emit("snip:remove", snip_id);
        },
    );

    socket.on(
        "snip:preview",
        async (SnipData: SnipData<BaseData, BaseView>) => {
            // Check if user has permission to write to the page
            if (!(socket.perms.pWrite === true)) {
                return socket.emit(
                    "error",
                    "You do not have access to write to this book. Could be finished!",
                );
            }

            socket
                .to("page-" + String(SnipData.page_id))
                .volatile.emit("snip:preview", SnipData);
        },
    );

    // Just a way to forward the event to the page namespace
    // and sync the placement of the snip
    socket.on("snip:placed", async (page_id: number, snip_id: number) => {
        // Check if user has permission to write to the page
        if (!(socket.perms.pWrite === true)) {
            return socket.emit(
                "error",
                "You do not have access to write to this book.",
            );
        }

        logWS(socket, "[Page: " + page_id + "] snip:placed " + snip_id);

        // Trigger page preview update
        await triggerRender(socket.book_id, page_id, server);
        // Remove placed snip
        book_nsp.emit("book:deleteQueuedSnip", snip_id);
    });
}

async function insert(
    { socket, book_nsp, server }: SocketData,
    snip_data: SnipData<BaseData, BaseView>,
    callback: (id: number | null) => void,
) {
    const page_id = snip_data.page_id;
    logWS(socket, "[Page: " + page_id + "] snip:insert");

    // Sanity check for page_id
    if (page_id === null || page_id === undefined) {
        console.error("[Socket][Book: " + socket.book_id + "] Page id is null");
        return;
    }

    // Sanity check book id
    snip_data.book_id = socket.book_id;
    snip_data.created_by = socket.perms.entity_id;

    // Sanity check to return incase of placement
    const old_id = snip_data.id;
    snip_data.id = null as unknown as number;

    const snip_data_new = await SQLService.snip.insert(snip_data);

    // Remove from preview stack if old_id is defined
    if (old_id && old_id !== -1) {
        server
            .of("/book-" + socket.book_id)
            .to("page-" + String(page_id))
            .emit("snip:preview:remove", old_id);
    }

    // Send snip to all other clients and insert into book/page
    socket.to("page-" + String(page_id)).emit("snip:insert", snip_data_new);

    // Send id back to initial client
    callback?.(snip_data_new.id);

    // Trigger page preview update
    await triggerRender(socket.book_id, page_id, server);
}

async function update(
    { socket, book_nsp, server }: SocketData,
    snip_data: SnipData<BaseData, BaseView>,
    callback: (id: number | null) => void,
) {
    const page_id = snip_data.page_id;

    // Sanity check for page_id
    if (page_id === null || page_id === undefined) {
        console.error("[Socket][Book: " + socket.book_id + "] Page id is null");
        return;
    }
    logWS(socket, "[Page: " + page_id + "] snip:update");

    // Sanity check book id
    snip_data.book_id = socket.book_id;
    snip_data.created_by = socket.perms.entity_id;

    // Sanity check if snip is in database
    let inserted_data;
    if (!snip_data.id || snip_data.id < 0) {
        inserted_data = await SQLService.snip.insert(snip_data);
    } else {
        //Check if in db
        const exists = await pool_data
            .query("SELECT id FROM snips WHERE id = ?", [snip_data.id])
            .then((res) => {
                return res.length > 0;
            });
        if (!exists) {
            inserted_data = await SQLService.snip.insert(snip_data);
            book_nsp
                .to("page-" + String(page_id))
                .emit("snip:preview:remove", snip_data.id);
        } else {
            inserted_data = await SQLService.snip.update(snip_data);
        }
    }

    // Send snip to all other clients and update in book/page
    const updated_snip = await SQLService.snip.getById(inserted_data.id);
    socket.to("page-" + String(page_id)).emit("snip:update", updated_snip);

    // Send id back to initial client
    callback?.(updated_snip.id);

    // Trigger page preview update
    await triggerRender(socket.book_id, page_id, server);
}
