"use server";
import { JwtPayload, sign, verify } from "jsonwebtoken";

import { ID } from "@snip/database/types";
import { config } from "@snip/utils/config";

export interface SocketTokenPayload extends JwtPayload {
    book_id: ID;
    user_id?: ID;
    token?: string;
}
export interface SocketTokenPayloadApi extends JwtPayload {
    api: boolean;
    date: number;
}

export function signToken(payload: SocketTokenPayload): string {
    return sign(payload, config.SOCKET.SECRET, {
        expiresIn: "24h",
    });
}

export function signApiToken(payload: SocketTokenPayloadApi): string {
    return sign(payload, config.SOCKET.SECRET, {
        expiresIn: "2h",
    });
}

/**
 * Verifies a socket token.
 * @param token - The token to verify.
 * @returns The payload of the verified token, including the book_id, user_id, and token.
 * @throws Error if the token is invalid or missing required properties.
 */
export function verifyToken(token: string): SocketTokenPayload {
    const payload = verify(token, config.SOCKET.SECRET);

    //Check if payload is valid
    if (!payload || typeof payload !== "object") {
        throw new Error("Invalid token" + JSON.stringify(payload));
    }

    // Check if book_id is present
    if (!payload.book_id) {
        throw new Error("Invalid token no book_id!");
    }
    // Either user_id or token must be present
    if (!payload.user_id && !payload.token) {
        throw new Error("Invalid token no user_id or token!");
    }
    // Add undefined to user_id and token
    if (!payload.user_id) {
        payload.user_id = undefined;
    }
    if (!payload.token) {
        payload.token = undefined;
    }

    return payload as SocketTokenPayload;
}

export function verifyApiToken(token: string): SocketTokenPayloadApi {
    const payload = verify(token, config.SOCKET.SECRET);

    //Check if payload is valid
    if (!payload || typeof payload !== "object") {
        throw new Error("Invalid token" + JSON.stringify(payload));
    }

    // Check if api is present
    if (!payload.api) {
        throw new Error("Invalid token no api!");
    }

    return payload as SocketTokenPayloadApi;
}
