import { registerService } from "@snip/utils";

import { check_mariadb_warning, NotFoundError } from "../errors";
import { pool_data as pool, pool_permissions } from "../sql.connection";
import { BookStrategy } from "../strategies";
import {
    BookData,
    BookDataResolved,
    BookOwner,
    Entity,
    ID,
    PartialBy,
    RESOURCE_BOOK,
} from "../types";
import permissionService from "./permission";

export class BookService implements BookStrategy {
    /**
     * Inserts a new book into the database.
     *
     * @param bookData - The data of the book to be inserted.
     * @returns A Promise that resolves to the inserted book data.
     */
    async insert(bookData: PartialBy<BookData, "id">): Promise<BookData> {
        const sql = `
        INSERT INTO books 
            (title, comment, owner_user_id, owner_group_id, cover_page_id ) 
        VALUES 
            (?, ?, ?, ?, ?);
    `;
        const bookData_id: ID = await pool
            .query(sql, [
                bookData.title,
                bookData.comment ? bookData.comment : null,
                bookData.owner_user_id ? bookData.owner_user_id : null,
                bookData.owner_group_id ? bookData.owner_group_id : null,
                bookData.cover_page_id ? bookData.cover_page_id : null,
            ])
            .then(check_mariadb_warning)
            .then((res) => {
                return res.insertId;
            });

        // Update the bookData with the new id
        bookData.id = bookData_id;
        bookData.last_updated = new Date();
        bookData.created = new Date();

        return bookData as BookData;
    }

    /**
     * Retrieves a book by its ID.
     *
     * @param book_id - The ID of the book to retrieve.
     * @returns A Promise that resolves to the BookData object representing the book.
     */
    async getById(book_id: ID): Promise<BookData> {
        const sql = "SELECT * FROM books WHERE id = ?";

        const bookData: BookData = await pool
            .query(sql, [book_id])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.length != 1) {
                    throw new NotFoundError("book", book_id);
                }
                return res[0];
            });

        return bookData;
    }

    /**
     * Retrieves an array of books by their IDs.
     * @param book_ids - An array of book IDs.
     * @returns A promise that resolves to an array of BookData objects.
     */
    async getByIds(book_ids: ID[]): Promise<BookData[]> {
        if (book_ids.length == 0) {
            return [];
        }

        const sql = "SELECT * FROM books WHERE id IN (?)";

        const books: BookData[] = await pool
            .query(sql, [book_ids])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.length != book_ids.length) {
                    console.warn(
                        `Missing books! Found '${res.length}' books for lookup of '
                        ${book_ids.length}' book ids.`,
                    );
                }
                return res;
            });

        return books;
    }

    /**
     * Retrieves a book by its ID from the resolved books table.
     * @param book_id - The ID of the book to retrieve.
     * @returns A promise that resolves to the retrieved book data.
     */
    async getByIdResolved(book_id: ID): Promise<BookDataResolved> {
        const sql = "SELECT * FROM books_resolved WHERE id = ?";
        const bookData: BookDataResolved = await pool
            .query(sql, [book_id])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.length != 1) {
                    throw new NotFoundError("book", book_id);
                }
                return res[0];
            });

        return bookData;
    }

    /**
     * Retrieves books by their IDs from the resolved books table.
     * @param book_ids - An array of book IDs.
     * @returns A promise that resolves to an array of BookDataResolved objects.
     */
    async getByIdsResolved(book_ids: ID[]): Promise<BookDataResolved[]> {
        if (book_ids.length == 0) {
            return [];
        }

        const sql = "SELECT * FROM books_resolved WHERE id IN (?)";

        const books: BookDataResolved[] = await pool
            .query(sql, [book_ids])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.length != book_ids.length) {
                    console.warn(
                        `Missing books! Found '${res.length}' books for lookup of '
                        ${book_ids.length}' book ids.`,
                    );
                }
                return res;
            });

        return books;
    }

    /**
     * Retrieves all books from the database.
     * @returns A promise that resolves to an array of BookData objects.
     */
    async getAll(): Promise<BookData[]> {
        const sql = "SELECT * FROM books";

        const books: BookData[] = await pool
            .query(sql)
            .then(check_mariadb_warning)
            .then((res) => {
                return res;
            });

        return books;
    }

    /**
     * Updates a book in the database.
     *
     * @param bookData - The data of the book to be updated.
     * @returns promise that resolves to true if the update was successful.
     */
    async update(bookData: BookData): Promise<BookData> {
        const sql = `
            UPDATE books SET 
                title=?, comment=?, owner_user_id=?, owner_group_id=?, cover_page_id=?
                WHERE id = ?;
            `;

        await pool
            .query(sql, [
                bookData.title,
                bookData.comment,
                bookData.owner_user_id,
                bookData.owner_group_id,
                bookData.cover_page_id,
                bookData.id,
            ])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.affectedRows == 0) {
                    throw new Error("Update failed, no rows affected");
                }
            });

        bookData.last_updated = new Date();

        return bookData;
    }

    /**
     * Deletes a book from the database.
     *
     * @param id - The ID of the book to delete.
     * @returns A promise that resolves to true if the deletion was successful.
     */
    async delete(id: number): Promise<boolean> {
        const sql = "DELETE FROM `books` WHERE id = ?";

        const success = await pool
            .query(sql, [id])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.affectedRows == 0) {
                    throw new Error("Delete failed, no rows affected");
                }
                return true;
            });

        return success;
    }

    /**
     * Sets the owner of a book entity.
     *
     * @param entity_id - The ID of the entity.
     * @param entity_type - The type of the entity (user or group).
     * @param book_id - The ID of the book.
     * @returns A promise that resolves to a boolean indicating whether the operation was successful.
     */
    async setOwner(
        entity_id: ID,
        entity_type: Entity | "user" | "group",
        book_id: ID,
    ): Promise<boolean> {
        let sql;
        switch (entity_type) {
            case "user":
            case 1:
                sql =
                    "UPDATE `books` SET `owner_user_id`=?, `owner_group_id`=NULL WHERE id = ?";
                break;
            case "group":
            case 2:
                sql =
                    "UPDATE `books` SET `owner_group_id`=?, `owner_user_id`=NULL WHERE id = ?";
                break;
        }

        // The perms table is updated by a trigger
        const success = await pool
            .query(sql, [entity_id, book_id])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.affectedRows == 0) {
                    throw new Error("Set owner failed, no rows affected");
                }
                return true;
            });

        return success;
    }

    /**
     * Sets the owner of a book by entity name.
     *
     * @param entity_name - The name of the entity (user or group).
     * @param book_id - The ID of the book.
     * @returns A Promise that resolves to the BookOwner object if successful, or null if there was an error or the entity does not exist.
     */
    async setOwnerByName(
        entity_name: string,
        book_id: number,
    ): Promise<BookOwner> {
        const sql_get_entity = `
            SELECT 
                id,
                "user" as entity_type,
                email as entity_name
            FROM users WHERE email = ?
            UNION
            SELECT
                id,
                "group" as entity_type,
                name as entity_name
            FROM groups WHERE name = ?
        `;
        const entity: BookOwner = await pool_permissions
            .query(sql_get_entity, [entity_name, entity_name])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.length != 1) {
                    throw new NotFoundError("entity", entity_name);
                }
                return res[0] as BookOwner;
            });

        // Transfer to new owner
        await this.setOwner(entity.id, entity.entity_type, book_id);

        return entity;
    }

    /**
     * Searches for books based on the given query.
     * @param query - The search query.
     * @returns A promise that resolves to an array of BookData objects.
     */
    async search(query: string): Promise<BookData[]> {
        const sql = `
        SELECT * FROM books 
            WHERE title LIKE ? OR
            id LIKE ? OR
            comment LIKE ? OR
            created LIKE ? OR
            finished LIKE ? OR
            last_updated LIKE ?
        `;

        const books: BookData[] = await pool
            .query(sql, [
                `%${query}%`,
                `%${query}%`,
                `%${query}%`,
                `%${query}%`,
                `%${query}%`,
                `%${query}%`,
            ])
            .then(check_mariadb_warning);

        return books;
    }

    /** Get all books by user id.
     * If pWrite is true, it returns only books where the user has write
     * permissions.
     *
     * @param user_id
     */
    async getByUserId(user_id: ID, pWrite = false): Promise<BookData[]> {
        const acl = await permissionService.getByEntity(user_id);
        const book_ids = acl
            .filter(
                (a) =>
                    a.resource_type == RESOURCE_BOOK &&
                    (pWrite ? a.pWrite : a.pRead),
            )
            .map((a) => a.resource_id);
        const books = await this.getByIdsResolved(book_ids);
        return books;
    }
}

/** We need to globally register each service here
 * to prevent memory leaks in development.
 */
export default registerService("db_book", () => new BookService());
