import { default as fs } from "fs";

import { pool_data, pool_permissions } from "@snip/database/sql.connection";
import { type fontsData } from "@snip/database/sql.typings";
import { config } from "@snip/utils/config";

import sqlts from "@rmp135/sql-ts";

const log = (...args: unknown[]) => console.log("[Prep]", ...args);

const COMMENT =
    "/* This file is generated by scripts/prepare.ts */\n/* eslint-disable */\n\n";

/** Script which is run before
 * the build command is executed.
 *
 * At the moment, this script is used to:
 * - update the version number
 * - generate the imports for the fonts
 * - generate the typings for the sql database
 */
async function prebuild() {
    await create_fonts_import();
    await create_sql_typings();
    await pool_permissions.end();
    await pool_data.end();
}

async function create_fonts_import() {
    const fonts = await pool_permissions
        .query<fontsData[]>("SELECT * FROM fonts")
        .then((rows) => {
            return rows;
        })
        .catch((err) => {
            console.error(err);
            return [];
        });

    // We need the fonts at two places, as import via css
    // and as import via js
    const file_js = __dirname + "/../database/src/fonts/internal.ts";
    let file_content = COMMENT;
    file_content += "export type AllowedFonts =";
    for (const font of fonts) {
        file_content += ` "${font.name}" |`;
    }
    file_content = file_content.slice(0, -1) + ";\n\n";

    file_content += "export const FONT_URLS: [string,AllowedFonts][] = [\n";
    for (const font of fonts) {
        file_content += `\t["${font.fname}", "${font.name}"],\n`;
    }
    file_content += "];\n\n";
    log(`Writing fonts mapping to ${file_js}`);
    fs.writeFileSync(file_js, file_content);

    const file_css =
        __dirname + "/../../apps/fullstack/components/styles/fonts.scss";
    let file_content_css = COMMENT;
    for (const font of fonts) {
        file_content_css += `@font-face {\n\tfont-family: "${font.name}";\n\tsrc: url("/fonts/${font.fname}");\n}\n`;
    }

    log(`Writing fonts mapping to ${file_css}`);
    fs.writeFileSync(file_css, file_content_css);
}

async function create_sql_typings() {
    const sql_ts_config = {
        client: "mysql",
        connection: {
            host: config.MYSQL.HOST,
            user: config.MYSQL.USER,
            password: config.MYSQL.PASSWORD,
            database: "snip_data",
        },
        interfaceNameFormat: "${table}Data", // User becomes UserData
        typeOverrides: {
            "snip_perms.acl.pRead": "boolean",
            "snip_perms.acl.pWrite": "boolean",
            "snip_perms.acl.pDelete": "boolean",
            "snip_perms.acl.pACL": "boolean",
            "snip_perms.users.emailValidated": "boolean",
            "snip_data.books_resolved.num_pages": "number",
        },
        typeMap: {
            tinyint: ["boolean"],
            enum: ["string"], // Add this line for enum type
            bigint: ["number"],
        },
        tableEnums: {
            "snip_data.background_types": {
                key: "name",
                value: "id",
            },
        },
    };
    // Get the SQL schema as typescript interfaces
    const data = await sqlts.toTypeScript(sql_ts_config);
    // Write the typescript interfaces to a file
    log("Writing SQL typings to ./packages/database/src/sql.typings.ts");
    fs.writeFileSync(__dirname + "/../database/src/sql.typings.ts", data);
}

log("Starting prebuild");
prebuild().then(() => {
    log("Finished prebuild");
});
